"""
Unit and regression test for the openxps package.

"""

# Import package, test suite, and other packages as needed
import sys


def test_openxps_imported():
    """Sample test, will always pass so long as import statement worked"""
    assert "openxps" in sys.modules
